package srctypes

const (
	DockerImageScheme     = "docker-image"
	DockerImageBlobScheme = "docker-image+blob"
	GitScheme             = "git"
	LocalScheme           = "local"
	HTTPScheme            = "http"
	HTTPSScheme           = "https"
	OCIScheme             = "oci-layout"
	OCIBlobScheme         = "oci-layout+blob"
)
